--[[
Library to track mobs spells casting/channeling in Wow Classic, (C) 2019 MiCHaeL released under MIT license.

Events: "SPELLCAST_START", "SPELLCAST_STOP", "SPELLCAST_FAILED" (this events are fired for channeling spells too)

Usage examples:

local lib = LibStub("LibKiwiClassicCast-1.0")
lib.Register('MyAddonName')
lib.Unregister('MyAddonName')
lib.RegisterEnabled('MyAddonName', boolean) -- Register|Unregister

-- example1: methods callbacks
function test:OnEvent(event, guid)
end
lib.RegisterCallback(guid, test, test.OnEvent)
lib.RegisterCallback(guid, test, 'OnEvent')
lib.UnregisterCallback(guid, test)

-- example2: functions callbacks
local function OnEvent(identifier, event, guid)
end
local identifier = string|table|function|...  --Use some addon_name prefix in "identifier" to avoid collisions with another addons.
lib.RegisterCallback(guid, identifier, OnEvent)
lib.UnregisterCallback(guid, identifier)
-- Retrieve casting or channeling info
local spellID, spellName, icon, timeStart, castDuration, isChanneling = lib.UnitCastingInfo(unit)
local spellID, spellName, icon, timeStart, castDuration, isChanneling = lib.GuidCastingInfo(guid)

This library borrows ideas and some spells lists from:
https://github.com/joev/LibClassicCast
https://github.com/wardz/ClassicCastbars
Thanks to joev & wardz
--]]

-- only for WoW Classic
if select(4, GetBuildInfo())>=20000 then return end

-- create Library
local lib = LibStub:NewLibrary("LibKiwiClassicCast-1.0",1)
if not lib then return end

-- fast access to some values
local next = next
local type = type
local tremove = table.remove
local GSI = function(id)
	if GetSpellInfo(id) then
		return GetSpellInfo(id)
	else
		return ""
	end
end
local GetTime = GetTime
local UnitGUID = UnitGUID
local CombatLogGetCurrentEventInfo = CombatLogGetCurrentEventInfo
local playerGUID = UnitGUID("player")

-- collect casted spellNames=>spellIDs
local castedSpells = lib.castedSpells
if not castedSpells then
	castedSpells = {}
	for i=1,35000 do
	   local name, _, _, dur = GSI(i)
	   if name and dur and dur>0 then
			 castedSpells[name] = i
		end
	end
	lib.castedSpells = castedSpells
end

-- channeled spells: spellName = {spellID, duration}
local channeledSpells = {
	[GSI(10)]    = {10, 7.5},	--Blizzard
	[GSI(605)]   = {605, 3},	--Mind Control
	[GSI(689)]   = {689,4.5},	--Drain Life
	[GSI(740)]   = {740,9.5},	--Tranquility
	[GSI(746)]   = {746,7},		--First Aid
	[GSI(755)]   = {755,10},	--Health Funnel
	[GSI(1002)]  = {1002,60},	--Eyes of the Beast
	[GSI(1120)]  = {1120,14.5},	--Drain Soul
	[GSI(1949)]  = {1949,15},	--Hellfire
	[GSI(2096)]  = {2096,60},	--Mind Vision
	[GSI(5138)]  = {5138,4.5},	--Drain Mana
	[GSI(5143)]  = {5143,4.5},	--Arcane Missiles
	[GSI(5740)]  = {5740,7.5},	--Rain of Fire
	[GSI(6197)]  = {6197,60},	--Eagle Eye
	[GSI(12051)] = {12051,8},	--Evocation
	[GSI(13278)] = {13278,4},	--Gnomish Death Ray
	[GSI(15407)] = {15407,3},	--Mind Flay
	[GSI(17401)] = {17401,9.5},	--Hurricane
	[GSI(20577)] = {20577,10},	--Cannibalize
}

-- crow control spells
local ccSpells = {
    [GSI(5211)] = true,       -- Bash
    [GSI(24394)] = true,      -- Intimidation
    [GSI(853)] = true,        -- Hammer of Justice
    [GSI(22703)] = true,      -- Inferno Effect (Summon Infernal)
    [GSI(408)] = true,        -- Kidney Shot
    [GSI(12809)] = true,      -- Concussion Blow
    [GSI(20253)] = true,      -- Intercept Stun
    [GSI(20549)] = true,      -- War Stomp
    [GSI(2637)] = true,       -- Hibernate
    [GSI(3355)] = true,       -- Freezing Trap
    [GSI(19386)] = true,      -- Wyvern Sting
    [GSI(118)] = true,        -- Polymorph
    [GSI(28271)] = true,      -- Polymorph: Turtle
    [GSI(28272)] = true,      -- Polymorph: Pig
    [GSI(20066)] = true,      -- Repentance
    [GSI(1776)] = true,       -- Gouge
    [GSI(6770)] = true,       -- Sap
    [GSI(1513)] = true,       -- Scare Beast
    [GSI(8122)] = true,       -- Psychic Scream
    [GSI(2094)] = true,       -- Blind
    [GSI(5782)] = true,       -- Fear
    [GSI(5484)] = true,       -- Howl of Terror
    [GSI(6358)] = true,       -- Seduction
    [GSI(5246)] = true,       -- Intimidating Shout
    [GSI(6789)] = true,       -- Death Coil
    [GSI(9005)] = true,       -- Pounce
    [GSI(1833)] = true,       -- Cheap Shot
    [GSI(16922)] = true,      -- Improved Starfire
    [GSI(19410)] = true,      -- Improved Concussive Shot
    [GSI(12355)] = true,      -- Impact
    [GSI(20170)] = true,      -- Seal of Justice Stun
    [GSI(15269)] = true,      -- Blackout
    [GSI(18093)] = true,      -- Pyroclasm
    [GSI(12798)] = true,      -- Revenge Stun
    [GSI(5530)] = true,       -- Mace Stun
    [GSI(19503)] = true,      -- Scatter Shot
    [GSI(605)] = true,        -- Mind Control
    [GSI(7922)] = true,       -- Charge Stun
    [GSI(18469)] = true,      -- Counterspell - Silenced
    [GSI(15487)] = true,      -- Silence
    [GSI(18425)] = true,      -- Kick - Silenced
    [GSI(24259)] = true,      -- Spell Lock
    [GSI(18498)] = true,      -- Shield Bash - Silenced
    [GSI(13327)] = true,      -- Reckless Charge
    [GSI(1090)] = true,       -- Sleep
    [GSI(5134)] = true,       -- Flash Bomb Fear
    [GSI(19821)] = true,      -- Arcane Bomb Silence
    [GSI(4068)] = true,       -- Iron Grenade
    [GSI(19769)] = true,      -- Thorium Grenade
    [GSI(13808)] = true,      -- M73 Frag Grenade
    [GSI(4069)] = true,       -- Big Iron Bomb
    [GSI(12543)] = true,      -- Hi-Explosive Bomb
    [GSI(4064)] = true,       -- Rough Copper Bomb
    [GSI(12421)] = true,      -- Mithril Frag Bomb
    [GSI(19784)] = true,      -- Dark Iron Bomb
    [GSI(4067)] = true,       -- Big Bronze Bomb
    [GSI(4066)] = true,       -- Small Bronze Bomb
    [GSI(4065)] = true,       -- Large Copper Bomb
    [GSI(13237)] = true,      -- Goblin Mortar
    [GSI(835)] = true,        -- Tidal Charm
    [GSI(13181)] = true,      -- Gnomish Mind Control Cap
    [GSI(12562)] = true,      -- The Big One
    [GSI(15283)] = true,      -- Stunning Blow (Weapon Proc)
    [GSI(56)] = true,         -- Stun (Weapon Proc)
    [GSI(26108)] = true,      -- Glimpse of Madness
}

-- fire callbacks events
local callbacks = {}

local function Fire(event, guid)
	local f = callbacks[guid]
	if f then
		for self,func in next,f do
			func(self, event, guid)
		end
	end
end

-- casts info
local castSID = {}
local castNam = {}
local castTex = {}
local castSta = {}
local castEnd = {}
local castCha = {}

local function StartCast(event, guid, spellID, spellName, texture, start, duration, channeled)
	if not start then
		start = GetTime()
	end
	if not (texture and duration) then
		local _, _, tex, dur = GSI(spellID)
		texture  = texture  or tex
		duration = duration or dur/1000
	end
	castSID[guid] = spellID
	castNam[guid] = spellName
	castTex[guid] = texture
	castSta[guid] = start
	castEnd[guid] = duration
	castCha[guid] = channeled
	if event then Fire(event, guid) end
end

local function EndCast(event, guid)
	castSID[guid] = nil
	if event then Fire(event, guid) end
end

-- combatlog event parsing
local frame = CreateFrame('Frame')
do
	local events = {
		SPELL_CAST_START = function(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
			local spellID = castedSpells[spellName]
			if spellID then
				StartCast('SPELLCAST_START', srcGUID, spellID, spellName)
			end
		end,
		SPELL_CAST_SUCCESS = function(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
			local data = channeledSpells[spellName]
			if data then
				StartCast('SPELLCAST_START', srcGUID, data[1], spellName, nil, nil, data[2], true) -- true => channeled spell
			elseif castSID[srcGUID] then
				EndCast('SPELLCAST_STOP', srcGUID)
			end
		end,
		SPELL_CAST_FAILED = function(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
			if castSID[srcGUID] and (srcGUID ~= playerGUID or not CastingInfo("player")) then -- Spamming spell keybinding triggers SPELL_CAST_FAILED, check if actually casting for the player
				EndCast('SPELLCAST_STOP', srcGUID)
			end
		end,
		SPELL_INTERRUPT = function(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
			if castSID[dstGUID] then
				EndCast('SPELLCAST_FAILED', dstGUID)
			end
		end,
		PARTY_KILL = function(srcGUID, srcFlags, dstGUID)
			castSID[dstGUID] = nil
		end,
		UNIT_DIED = function(srcGUID, srcFlags, dstGUID)
			castSID[dstGUID] = nil
		end,
		SPELL_AURA_APPLIED = function(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
			if ccSpells[spellName] and castSID[dstGUID] then
				EndCast('SPELLCAST_FAILED', dstGUID)
			end
		end,
		SPELL_AURA_REMOVED = function(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
			local data = channeledSpells[spellName]
			if data and castSID[srcGUID] then
				EndCast('SPELLCAST_STOP', srcGUID)
			end
		end,
 	}
	frame:SetScript("OnEvent", function()
		local _, event, _, srcGUID, _, srcFlags, _, dstGUID,  _, dstFlags, _, _, spellName  = CombatLogGetCurrentEventInfo()
		local func = events[event]
		if func then
			func(srcGUID, srcFlags, dstGUID, dstFlags, spellName)
		end
	end)
end

-- public methods
local register = {}

local function Register(addonName)
	if not next(register) then
		frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
	register[addonName] = true
end

local function Unregister(addonName)
	register[addonName] = nil
	if not next(register) then
		frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

local function RegisterEnabled(addonName, enabled)
	if enabled then
		Register(addonName)
	else
		Unregister(addonName)
	end
end

local tables = {}

local function RegisterCallback(guid, self, func)
	local f = callbacks[guid]
	if not f then
		f = tremove(tables) or {}; callbacks[guid] = f
	end
	f[self] = type(func)=='function' and func or self[func]
end

local function UnregisterCallback(guid, self)
	local f = callbacks[guid]
	if f then
		f[self] = nil
		if not next(f) then
			tables[#tables+1] = f
			callbacks[guid] = nil
		end
	end
end

local function GuidCastingInfo(guid)
	spellID = castSID[guid]
	if spellID then
		return spellID, castNam[guid], castTex[guid], castSta[guid], castEnd[guid], castCha[guid]
	end
end

local function UnitCastingInfo(unit)
	local guid = UnitGUID(unit)
	if guid then
		GuidCastingInfo(guid)
	end
end

lib.Register           = Register
lib.Unregister         = Unregister
lib.RegisterEnabled	   = RegisterEnabled
lib.RegisterCallback   = RegisterCallback
lib.UnregisterCallback = UnregisterCallback
lib.GuidCastingInfo    = GuidCastingInfo
lib.UnitCastingInfo    = UnitCastintInfo
